use std::collections::HashMap;
use std::convert::Infallible;
use std::sync::{Arc, Mutex};
use std::thread;
use std::time::Duration;

use anyhow::{anyhow, Result};
use easee::api::{self, ChargerOpMode, Circuit, Context, SetCurrent, Triphase};
use easee::observation::{self, Event, ObservationError};
use tracing::{error, info, warn};

use crate::config::{self, Config};
use crate::mattermost::{self, Channel};
use crate::prom::PromClient;

use observation::{Observation, PilotMode};

struct Charger {
    inner: api::Charger,
    owners: Vec<String>,
    current: Option<(f64, f64, f64)>,
}

impl Charger {
    pub fn from_api(inner: api::Charger, configs: &[config::Charger]) -> Self {
        let owners = configs
            .iter()
            .find(|c| c.id == inner.id)
            .map(|c| &c.owners)
            .cloned()
            .unwrap_or_default();
        Charger {
            inner,
            owners,
            current: None,
        }
    }
}

pub fn start(
    mut ctx: Context,
    config: Config,
    mut chargers: Vec<api::Charger>,
) -> Result<Infallible> {
    let mattermost = mattermost::Context::new(config.mattermost.base, &config.mattermost.token)?;
    let mut stream = observation::Stream::from_context(&mut ctx)?;

    let chargers: HashMap<String, Charger> = chargers
        .into_iter()
        .filter_map(|c| {
            stream
                .subscribe(&c.id)
                .map_err(|e| error!("Cannot subscribe {}: {e}", &c.id))
                .ok()?;
            let name = c.name.clone();
            c.enable_smart_charging(&mut ctx)
                .map_err(|e| error!("Cannot enable smart charging on {}: {e}", &c.id))
                .ok()?;
            Some((name, Charger::from_api(c, &config.chargers)))
        })
        .collect();

    let chargers = Arc::new(Mutex::new(chargers));

    // TODO get channel on a per-charger basis
    let channel = mattermost.channel("9d9o1a5qf7fofk3wqfa493gkfe");

    info!("Controller started");
    mattermost.send_to_channel(&channel, "Easee Controller started")?;

    if let Some(reg) = config.regulator {
        let circuit = ctx
            .sites()?
            .into_iter()
            .find(|s| s.id == reg.site_id)
            .ok_or(anyhow!("Invalid site id {}", reg.site_id))?
            .details(&mut ctx)?
            .circuits
            .into_iter()
            .find(|c| c.id == reg.circuit_id)
            .ok_or(anyhow!("Invalid circuit id {}", reg.circuit_id))?;

        let controller = Controller {
            bias: Triphase::from(reg.power_bias_watts),
            ctx,
            prom: PromClient::new(config.prometheus.base),
            circuit,
            delay: Duration::from_secs(reg.polling_interval),
            p: reg.p,
            i: reg.i,
            d: reg.d,
        };

        let _ctrl = thread::spawn(move || controller.adjust_power());
    }

    loop {
        let evt = match stream.recv() {
            Ok(e) => e,
            Err(ObservationError::Stream(stream_error)) => Err(stream_error)?,
            Err(other) => {
                error!("Cannot process message: {}", other);
                continue;
            }
        };

        let mut chargers = chargers.lock().unwrap();

        let Some(charger) = chargers.get_mut(&evt.charger) else {
            warn!("Received message for unknown charger {}", &evt.charger);
            continue;
        };

        let result = handle_event(evt, charger, &mattermost, &channel);

        if let Err(err) = result {
            error!("Error handling observation: {:?}", err);
        }
    }
}

fn handle_event(
    evt: Event,
    charger: &mut Charger,
    ctx: &mattermost::Context,
    channel: &Channel,
) -> Result<()> {
    let send = |msg: &str| ctx.send_to_channel(channel, msg);

    match evt.observation {
        Observation::PilotMode(mode) => match mode {
            PilotMode::Disconnected => send("Car Disconnected"),
            PilotMode::Connected => send("Car Connected"),
            PilotMode::Charging => send("Car Charging"),
            PilotMode::NeedsVentilation => send("Car needs ventilation"),
            PilotMode::FaultDetected => send("Fault detected"),
            PilotMode::Unknown => send("Unknown"),
        },
        Observation::ChargerOpMode(mode) => {
            match mode {
                ChargerOpMode::Unknown => send("Unknown"),
                ChargerOpMode::Disconnected => send("Charger disconnected"),
                ChargerOpMode::Paused => send("Charge paused"),
                ChargerOpMode::Charging => send("Charging"),
                ChargerOpMode::Finished => send("Charging finished"),
                ChargerOpMode::Error => send("Charger error"),
                ChargerOpMode::Ready => send("Charger ready"),
                ChargerOpMode::AwaitingAuthentication => send("Charger awaiting authentication"),
                ChargerOpMode::Deauthenticating => send("Charger deauthenticating"),
            }?;
            ctx.set_status(mode)
        }
        other => Ok(info!("{}: {:?}", evt.charger, other)),
    }
}

struct Controller {
    /// Power bias in W
    bias: Triphase,
    ctx: Context,
    prom: PromClient,
    circuit: Circuit,
    delay: Duration,
    p: f64,
    i: f64,
    d: f64,
}

impl Controller {
    pub fn adjust_power(mut self) -> Result<()> {
        let voltage = 240f64;
        let time_to_live = Some(self.delay.as_secs() as i32);

        let mut current = self.circuit.dynamic_current(&mut self.ctx)?;
        let mut integrated = Triphase::default();
        let mut prev_available_current = Triphase::default();

        loop {
            let export_power = self.prom.current_power()?;
            let available_power = export_power - self.bias;
            let available_current = available_power * (1.0 / voltage);

            integrated = available_current * 0.3 + integrated * (1.0 - 0.3);
            let differentiated = available_current - prev_available_current;
            prev_available_current = available_current;

            let delta = available_current * self.p + differentiated * self.d + integrated * self.i;

            current = current + delta;
            self.circuit.set_dynamic_current(
                &mut self.ctx,
                SetCurrent {
                    time_to_live,
                    current,
                },
            )?;

            thread::sleep(self.delay);
        }
    }
}
